@@ -1,10 +1,13 @@
+from typing import Optional
 import warnings
 import inspect
+import copy
 
 import torch
 
 
 _E3NN_COMPILE_MODE = "__e3nn_compile_mode__"
+_VALID_MODES = ('trace', 'script', 'unsupported', None)
 _MAKE_TRACING_INPUTS = '_make_tracing_inputs'
 
 
@@ -16,16 +19,15 @@ def compile_mode(mode: str):
         mode : str
             'script', 'trace', or None
     """
-    if mode not in ['trace', 'script', None]:
+    if mode not in _VALID_MODES:
         raise ValueError("Invalid compile mode")
+
     def decorator(obj):
         if not (inspect.isclass(obj) and issubclass(obj, torch.nn.Module)):
             raise TypeError("@e3nn.util.jit.compile_mode can only decorate classes derived from torch.nn.Module")
-        if hasattr(obj, _E3NN_COMPILE_MODE):
-            # TODO: is this right for subclasses?
-            warnings.warn("Something is strange — did this class get marked twice with @e3nn.util.jit.compile_mode?")
         setattr(obj, _E3NN_COMPILE_MODE, mode)
         return obj
+
     return decorator
 
 
@@ -44,7 +46,7 @@ def get_compile_mode(mod: torch.nn.Module) -> str:
         mode = getattr(mod, _E3NN_COMPILE_MODE)
     else:
         mode = getattr(type(mod), _E3NN_COMPILE_MODE, None)
-    assert mode in ['script', 'trace', None], "Invalid compile mode `%r`" % mode
+    assert mode in _VALID_MODES, "Invalid compile mode `%r`" % mode
     return mode
 
 
@@ -53,6 +55,7 @@ def compile(
     n_trace_checks: int = 1,
     script_options: dict = {},
     trace_options: dict = {},
+    in_place: bool = True,
 ):
     """Recursively compile a module and all submodules according to their decorators.
 
@@ -73,6 +76,12 @@ def compile(
     -------
     Returns the compiled module.
     """
+    mode = get_compile_mode(mod)
+    if mode == 'unsupported':
+        raise NotImplementedError(f"{type(mod).__name__} does not support TorchScript compilation")
+
+    if not in_place:
+        mod = copy.deepcopy(mod)
     # TODO: debug logging
     assert n_trace_checks >= 1
     # == recurse to children ==
@@ -85,18 +94,20 @@ def compile(
                 submod,
                 n_trace_checks=n_trace_checks,
                 script_options=script_options,
-                trace_options=trace_options
+                trace_options=trace_options,
+                in_place=True,  # since we deepcopied the module above, we can do inplace
             )
         )
     # == Compile this module now ==
-    mode = get_compile_mode(mod)
-
     if mode == 'script':
         mod = torch.jit.script(mod, **script_options)
     elif mode == 'trace':
         # These are always modules, so we're always using trace_module
         # We need tracing inputs:
-        check_inputs = get_tracing_inputs(mod, n_trace_checks)
+        check_inputs = get_tracing_inputs(
+            mod,
+            n_trace_checks,
+        )
         assert len(check_inputs) >= 1, "Must have at least one tracing input."
         # Do the actual trace
         mod = torch.jit.trace_module(
@@ -108,7 +119,11 @@ def compile(
     return mod
 
 
-def get_tracing_inputs(mod: torch.nn.Module, n: int = 1):
+def get_tracing_inputs(
+    mod: torch.nn.Module,
+    n: int = 1,
+    device: Optional[torch.device] = None
+):
     """Get random tracing inputs for ``mod``.
 
     First checks if ``mod`` has a ``_make_tracing_inputs`` method. If so, calls it with ``n`` as the single argument and returns its results.
@@ -120,6 +135,8 @@ def get_tracing_inputs(mod: torch.nn.Module, n: int = 1):
         mod : torch.nn.Module
         n : int, default = 1
             A hint for how many inputs are wanted. Usually n will be returned, but modules don't necessarily have to.
+        device : torch.device
+            The device to do tracing on.
 
     Returns
     -------
@@ -127,7 +144,7 @@ def get_tracing_inputs(mod: torch.nn.Module, n: int = 1):
         Tracing inputs in the format of ``torch.jit.trace_module``: dicts mapping method names like ``'forward'`` to tuples of arguments.
     """
     # Avoid circular imports
-    from ._argtools import _get_io_irreps, _rand_args, _to_device
+    from ._argtools import _get_io_irreps, _rand_args, _to_device, _get_device
     # - Get inputs -
     if hasattr(mod, _MAKE_TRACING_INPUTS):
         # This returns a trace_module style dict of method names to test inputs
@@ -144,12 +161,8 @@ def get_tracing_inputs(mod: torch.nn.Module, n: int = 1):
         )
         trace_inputs = [{'forward': _rand_args(irreps_in)} for _ in range(n)]
     # - Put them on the right device -
-    # Try to a get a parameter
-    a_buf = next(mod.parameters(), None)
-    if a_buf is None:
-        # If there isn't one, try to get a buffer
-        a_buf = next(mod.buffers(), None)
-    device = a_buf.device if a_buf is not None else 'cpu'
+    if device is None:
+        device = _get_device(mod)
     # Move them
     trace_inputs = _to_device(trace_inputs, device)
     return trace_inputs
@@ -159,6 +172,7 @@ def trace_module(
     mod: torch.nn.Module,
     inputs: dict = None,
     check_inputs: list = [],
+    in_place: bool = True
 ):
     """Trace a module.
 
@@ -190,7 +204,7 @@ def trace_module(
         )
 
     # Compile
-    out = compile(mod)
+    out = compile(mod, in_place=in_place)
 
     # Restore old values, if we had them
     if old_mode is not None:
@@ -204,6 +218,7 @@ def trace(
     mod: torch.nn.Module,
     example_inputs: tuple = None,
     check_inputs: list = [],
+    in_place: bool = True
 ):
     """Trace a module.
 
@@ -222,10 +237,11 @@ def trace(
         mod=mod,
         inputs=({'forward': example_inputs} if example_inputs is not None else None),
         check_inputs=[{'forward': c} for c in check_inputs],
+        in_place=in_place
     )
 
 
-def script(mod: torch.nn.Module):
+def script(mod: torch.nn.Module, in_place: bool = True):
     """Script a module.
 
     Like ``torch.jit.script``, but first recursively compiles ``mod`` using :func:``compile``.
@@ -244,7 +260,7 @@ def script(mod: torch.nn.Module):
     setattr(mod, _E3NN_COMPILE_MODE, 'script')
 
     # Compile
-    out = compile(mod)
+    out = compile(mod, in_place=in_place)
 
     # Restore old values, if we had them
     if old_mode is not None:
